/*
 * Copyright 2009 SIB Visions GmbH
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 *
 *
 * History
 *
 * 17.11.2008 - [HM] - creation
 * 26.06.2009 - [JR] - remove: index < size - 1 --> index < size [BUGFIX]
 * 04.08.2009 - [JR] - set/isNavigationKeysEnabled implemented
 * 24.10.2012 - [JR] - #604: added constructor
 */
package javax.rad.genui.container;

import javax.rad.genui.UIContainer;
import javax.rad.genui.UIFactoryManager;
import javax.rad.genui.UIImage;
import javax.rad.ui.IComponent;
import javax.rad.ui.IImage;
import javax.rad.ui.container.ITabsetPanel;
import javax.rad.ui.event.TabsetHandler;
import javax.rad.ui.event.UITabsetEvent;

import com.sibvisions.util.ArrayUtil;

/**
 * Platform and technology independent TabSetPanel.
 * It is designed for use with AWT, Swing, SWT, JSP, JSF, ... .
 * 
 * @author Martin Handsteiner
 */
public class UITabsetPanel extends UIContainer<ITabsetPanel> 
                           implements ITabsetPanel
{
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Class members
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/** images of the tabs. */
	private ArrayUtil<IImage> images = new ArrayUtil<IImage>();
	
	/** text of the tabs. */
	private ArrayUtil<String> texts = new ArrayUtil<String>();

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Initialization
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	  
    /**
     * Creates a new instance of <code>UITabsetPanel</code>.
     *
     * @see ITabsetPanel
     */
	public UITabsetPanel()
	{
		this(UIFactoryManager.getFactory().createTabsetPanel());
	}

    /**
     * Creates a new instance of <code>UITabsetPanel</code> with the given
     * tabset panel.
     *
     * @param pPanel the tabset panel
     * @see ITabsetPanel
     */
	protected UITabsetPanel(ITabsetPanel pPanel)
	{
		super(pPanel);
		
		uiResource.eventTabClosed().addListener(this, "doTabClosed");
		uiResource.eventTabMoved().addListener(this, "doTabMoved");
	}
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Interface implementation
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * {@inheritDoc}
	 */
	public void setTabPlacement(int pPlacement)
    {
    	uiResource.setTabPlacement(pPlacement);
    }
	
	/**
	 * {@inheritDoc}
	 */
	public int getTabPlacement()
    {
    	return uiResource.getTabPlacement();
    }
	
	/**
	 * {@inheritDoc}
	 */
	public void setEnabledAt(int pTabPosition, boolean pEnabled)
    {
    	uiResource.setEnabledAt(pTabPosition, pEnabled);
    }
	
	/**
	 * {@inheritDoc}
	 */
	public boolean isEnabledAt(int pTabPosition)
    {
    	return uiResource.isEnabledAt(pTabPosition);
    }
	
	/**
	 * {@inheritDoc}
	 */
	public void setTabLayoutPolicy(int pLayoutPolicy)
    {
    	uiResource.setTabLayoutPolicy(pLayoutPolicy);
    }
	
	/**
	 * {@inheritDoc}
	 */
	public int getTabLayoutPolicy()
    {
    	return uiResource.getTabLayoutPolicy();
    }
	
	/**
	 * {@inheritDoc}
	 */
	public void setSelectedIndex(int pIndex)
    {
    	uiResource.setSelectedIndex(pIndex);
    }
	
	/**
	 * {@inheritDoc}
	 */
	public int getSelectedIndex()
    {
    	return uiResource.getSelectedIndex();
    }

	/**
	 * {@inheritDoc}
	 */
	public IImage getIconAt(int pIndex)
	{
		return images.get(pIndex);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void setIconAt(int pIndex, IImage pImage)
	{
		// ensure that the factory gets his own resource, to prevent exception on factory internal casts!
		if (pImage instanceof UIImage)
		{
	    	uiResource.setIconAt(pIndex, ((UIImage)pImage).getUIResource());
		}
		else
		{
	    	uiResource.setIconAt(pIndex, pImage);
		}

		images.set(pIndex, pImage);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void setClosableAt(int pTabPosition, boolean pClosable)
    {
    	uiResource.setClosableAt(pTabPosition, pClosable);
    }
	
	/**
	 * {@inheritDoc}
	 */
	public boolean isClosableAt(int pTabPosition)
    {
    	return uiResource.isClosableAt(pTabPosition);
    }
	
	/**
	 * {@inheritDoc}
	 */
	public void setDragable(boolean pDragable)
    {
    	uiResource.setDragable(pDragable);
    }
	
	/**
	 * {@inheritDoc}
	 */
	public boolean isDragable()
    {
    	return uiResource.isDragable();
    }
	
	/**
	 * {@inheritDoc}
	 */
	public void setTextAt(int pIndex, String pText)
	{
		uiResource.setTextAt(pIndex, translate(pText));
		
		texts.set(pIndex, pText);
	}

	/**
	 * {@inheritDoc}
	 */
	public String getTextAt(int pIndex)
	{
		return texts.get(pIndex);
	}
	
	/**
	 * {@inheritDoc}
	 */
	public void setNavigationKeysEnabled(boolean pEnabled)
	{
		uiResource.setNavigationKeysEnabled(pEnabled);
	}

	/**
	 * {@inheritDoc}
	 */
	public boolean isNavigationKeysEnabled()
	{
		return uiResource.isNavigationKeysEnabled();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public TabsetHandler eventTabClosed()
	{
		return uiResource.eventTabClosed();
	}
	
	/**
	 * {@inheritDoc}
	 */
	public TabsetHandler eventTabMoved()
	{
		return uiResource.eventTabMoved();
	}
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Overwritten methods
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void add(IComponent pComponent, Object pConstraints, int pIndex)
	{
		super.add(pComponent, translate((String)pConstraints), pIndex);
		
		if (pIndex < 0)
		{
			images.add(null);
			texts.add((String)pConstraints);
		}
		else
		{
			images.add(pIndex, null);
			texts.add(pIndex, (String)pConstraints);
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	@Override
	public void remove(int pIndex)
	{
		super.remove(pIndex);

		images.remove(pIndex);
		texts.remove(pIndex);
	}

	/**
	 * {@inheritDoc}
	 */
	@Override
	public void updateTranslation()
	{
		boolean bChanged = isTranslationChanged();

		super.updateTranslation();

		if (bChanged)
		{
			//Update tab texts
			for (int i = 0, anz = texts.size(); i < anz; i++)
			{
				uiResource.setTextAt(i, translate(texts.get(i)));
			}
		}
	}	
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// User-defined methods
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * Notification that a tab was closed.
	 * 
	 * @param pEvent the event from the tabset
	 */
	public void doTabClosed(UITabsetEvent pEvent)
	{
		int iIndex = pEvent.getOldIndex();
		
		remove(iIndex);
	}
	
	/**
	 * Notfication that a tab was moved.
	 * 
	 * @param pEvent the event from the tabset
	 */
	public void doTabMoved(UITabsetEvent pEvent)
	{
		int iOld = pEvent.getOldIndex();
		int iNew = pEvent.getNewIndex();

		IComponent comp = getComponent(iOld);
		String text  = texts.get(iOld);
		IImage image = images.get(iOld);

		remove(iOld);
		
		add(comp, text, iNew);
		setIconAt(iNew, image);
	}
	
}	// UITabsetPanel
