/*
 * Copyright 2009 SIB Visions GmbH
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 *
 *
 * History
 *
 * 22.04.2008 - [HM] - creation
 * 17.07.2009 - [JR] - getLength, getInputStream: used fileName property as option
 * 05.10.2010 - [JR] - #169: remove directories from filenames
 * 11.10.2011 - [JR] - #480: always put this into the cache
 */
package javax.rad.io;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.Serializable;
import java.net.MalformedURLException;
import java.net.URL;

import com.sibvisions.util.ObjectCache;
import com.sibvisions.util.type.FileUtil;

/**
 * RemoteFileHandle handles data that is stored in the ObjectCache.
 * Supported Objects are java.io.File, java.net.URL, javax.rad.io.IFileHandle and byte[]. 
 * The Bean standard guarantees with the properties "fileName" and "objectCacheKey" that it
 * can be transported with a connection.
 * The Creator of the ObjectCacheHandle has to guarantee that the data is available under the given key.
 * 
 * For compatibility reasons is the ObjectCacheHandle also Serializable.
 * 
 * @author Martin Handsteiner
 */
public class RemoteFileHandle implements IFileHandle, 
                                         Serializable
{
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Class members
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/** Placeholder for the ObjectCache key. */
	public static final String OBJECT_CACHE_KEY = "OBJECT_CACHE_KEY";

	/** The default cache timeout. */
	private static final long DEFAULT_TIMEOUT = 300000;

	/** The url where to download the content with key. */
	private static String downloadURLSpec = null;
	
	/** The content. */ 
	private String fileName;

	/** The key to find the content. */
	private Object objectCacheKey;
	
	/** The current cache timeout. */
	private long timeout = DEFAULT_TIMEOUT;
	
	/** The file reference. */
	private transient File fiTempOut = null;
	
	/** the content file. */
	private transient File fiContent = null;
	
	/** the content URL. */
	private transient URL urlContent = null;
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Initialization
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * Constructs a new FileHandle.
	 */
	public RemoteFileHandle()
	{
		fileName = null;
		objectCacheKey = null;
	}
	
	/**
	 * Constructs a new FileHandle.
	 * @param pFileName the file name.
	 */
	public RemoteFileHandle(String pFileName)
	{
		fileName = FileUtil.getName(pFileName);
		objectCacheKey = null;
	}
	
	/**
	 * Constructs a new FileHandle. 
	 * @param pFileName the filename.
	 * @param pObjectCacheKey the ObjectCache key.
	 */
	public RemoteFileHandle(String pFileName, Object pObjectCacheKey)
	{
		fileName = FileUtil.getName(pFileName);
		objectCacheKey = pObjectCacheKey;
	}
	
	/**
	 * Constructs a new FileHandle.
	 * @param pFileName the filename.
	 * @param pContent the content.
	 */
	public RemoteFileHandle(String pFileName, URL pContent)
	{
		this(pFileName, null, pContent);
	}
	
	/**
	 * Constructs a new FileHandle.
	 * @param pFileName the filename.
	 * @param pObjectCacheKey the ObjectCache key.
	 * @param pContent the content.
	 */
	public RemoteFileHandle(String pFileName, Object pObjectCacheKey, URL pContent)
	{
		fileName = FileUtil.getName(pFileName);
		objectCacheKey = pObjectCacheKey;
		setContent(pContent);
	}
	
	/**
	 * Constructs a new FileHandle.
	 * @param pFileName the filename.
	 * @param pContent the content.
	 * @throws IOException if an IOException occurs.
	 */
	public RemoteFileHandle(String pFileName, InputStream pContent) throws IOException
	{
		this(pFileName, null, pContent);
	}
	
	/**
	 * Constructs a new FileHandle.
	 * @param pFileName the filename.
	 * @param pObjectCacheKey the ObjectCache key.
	 * @param pContent the content.
	 * @throws IOException if an IOException occurs.
	 */
	public RemoteFileHandle(String pFileName, Object pObjectCacheKey, InputStream pContent) throws IOException
	{
		fileName = FileUtil.getName(pFileName);
		objectCacheKey = pObjectCacheKey;
		setContent(pContent);
	}
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Interface implementation
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * {@inheritDoc}
	 */
	public String getFileName()
	{
		return fileName;
	}

	/**
	 * {@inheritDoc}
	 */
	public InputStream getInputStream() throws IOException
	{
		if (fiTempOut != null)
		{
			return new FileInputStream(fiTempOut);
		}
		else if (objectCacheKey != null)
		{
			Object content = ObjectCache.get(objectCacheKey);

			if (content == null)
			{
				String url = getDownloadURL();
				if (url != null)
				{
					return new URL(url).openConnection().getInputStream();
				}
			}
			else if (content instanceof IFileHandle)
			{
				if (content != this)
				{
					return ((IFileHandle)content).getInputStream();
				}
				else
				{
					if (fiContent != null)
					{
						return new FileInputStream(fiContent);
					}
					else if (urlContent != null)
					{
						return urlContent.openConnection().getInputStream();
					}
				}
			}
		}
		
		return null;
	}
	
	/**
	 * {@inheritDoc}
	 */
	public long getLength() throws IOException
	{
		if (fiTempOut != null)
		{
			return fiTempOut.length();
		}
		else if (objectCacheKey != null)
		{
			Object content = ObjectCache.get(objectCacheKey);

			if (content == null)
			{
				String url = getDownloadURL();
				if (url != null)
				{
					return new URL(url).openConnection().getContentLength();
				}
			}
			else if (content instanceof IFileHandle)
			{
				if (content != this)
				{
					return ((IFileHandle)content).getLength();
				}
				else
				{
					if (fiContent != null)
					{
						return fiContent.length();
					}
					else if (urlContent != null)
					{
						return urlContent.openConnection().getContentLength();
					}
				}
			}
		}
		
		return -1;
	}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// User-defined methods
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * Gets the download URL, that specifies where to get the content from server.
	 * If the URL contains OBJECT_CACHE_KEY the ObjectCache key is filled in at that place, 
	 * else is the key concatenated to the end. 
	 * 
	 * @return the download URL, that specifies where to get the content from server.
	 */
	public static String getDownloadURLSpec()
	{
		return downloadURLSpec;
	}

	/**
	 * Sets the download URL, that specifies where to get the content from server.
	 * If the URL contains a OBJECT_CACHE_KEY the ObjectCache key is filled in at that place, 
	 * else is the key concatinated to the end. 
	 *
	 * @param pDownloadURLSpec the download URL, that specifies where to get the content from server.
	 */
	public static void setDownloadURLSpec(String pDownloadURLSpec)
	{
		downloadURLSpec = pDownloadURLSpec;
	}

	/**
	 * Creates the download URL with filled in key.
	 * 
	 * @param pObjectCacheKey the ObjectCache key.
	 * @return the download URL with filled in key.
	 * @throws MalformedURLException if the download url is a malformed.
	 */
	public static String getDownloadURL(Object pObjectCacheKey) throws MalformedURLException
	{
		int index = downloadURLSpec.indexOf(OBJECT_CACHE_KEY);
		if (index < 0)
		{
			return downloadURLSpec + pObjectCacheKey;
		}
		else
		{
			return downloadURLSpec.substring(0, index) + 
			       pObjectCacheKey + 
			       downloadURLSpec.substring(index + OBJECT_CACHE_KEY.length());
		}
	}
	
	/**
	 * Gets the download URL with filled in key.
	 * 
	 * @return the download URL with filled in key.
	 * @throws MalformedURLException if the download url is a malformed.
	 */
	public String getDownloadURL() throws MalformedURLException
	{
		if (downloadURLSpec == null || objectCacheKey == null)
		{
			return null;
		}
		else
		{
			return getDownloadURL(objectCacheKey);
		}
	}
	
	/**
	 * Gets the access key for the object.
	 * 
	 * @return the access key for the object                 
	 */
	public static Object createObjectCacheKey()
	{
		return ObjectCache.createKey();
	}
	
	/**
	 * Sets the file name of this file handle.
	 * @param pFileName the file name of this file handle.
	 */
	public void setFileName(String pFileName)
	{
		fileName = FileUtil.getName(pFileName);
	}

	/**
	 * Gets the ObjectCache key. 
	 * @return the ObjectCache key. 
	 */
	public Object getObjectCacheKey()
	{
		return objectCacheKey;
	}
	
	/**
	 * Sets the ObjectCache key. 
	 * @param pObjectCacheKey the ObjectCache key. 
	 */
	public void setObjectCacheKey(Object pObjectCacheKey)
	{
		objectCacheKey = pObjectCacheKey;
	}

	/**
	 * Gets the ObjectCache key. 
	 * @return the ObjectCache key. 
	 */
	public long getTimeout()
	{
		return timeout;
	}
	
	/**
	 * Sets the ObjectCache key. 
	 * @param pTimeout the ObjectCache key. 
	 */
	public void setTimeout(long pTimeout)
	{
		timeout = pTimeout;
	}

	/**
	 * Resets the content to zero, and gets an OutputStresm to write the content.
	 * The content is stored temporary with the configured timeout.
	 * @return an OutputStresm to write the content.
	 * @throws IOException if an IOException occurs.
	 */
	public OutputStream getOutputStream() throws IOException
	{
		fiTempOut = File.createTempFile(getClass().getSimpleName(), ".tmp");
		fiTempOut.deleteOnExit();
		
		if (objectCacheKey == null)
		{
			objectCacheKey = ObjectCache.put(this, timeout);
		}
		else
		{
			ObjectCache.put(objectCacheKey, this, timeout);
		}
		
		return new FileOutputStream(fiTempOut);
	}

	/**
	 * Sets a new content with a byte array.
	 * The content is stored temporary with the configured timeout.
	 * @param pContent the new content.
	 * @throws IOException if an IOException occurs.
	 */
	public void setContent(byte[] pContent) throws IOException
	{
		if (pContent == null)
		{
			if (objectCacheKey != null)
			{
				ObjectCache.remove(objectCacheKey);
			}
		}
		else
		{  
			OutputStream outStream = getOutputStream();
			outStream.write(pContent);
			outStream.close();
		}
	}
	
	/**
	 * Sets a new content with a InputStream.
	 * The content is stored temporary with the configured timeout.
	 * @param pContent the new content.
	 * @throws IOException if an IOException occurs.
	 */
	public void setContent(InputStream pContent) throws IOException
	{
		if (pContent == null)
		{
			if (objectCacheKey != null)
			{
				ObjectCache.remove(objectCacheKey);
			}
		}
		else
		{
			FileUtil.copy(pContent, true, getOutputStream(), true);
		}
	}
	
	/**
	 * Sets a new content with a File.
	 * The content is stored temporary with the configured timeout.
	 * @param pContent the new content.
	 */
	public void setContent(File pContent)
	{
		urlContent = null;
		fiContent = pContent;
		
		if (pContent == null)
		{
			if (objectCacheKey != null)
			{
				ObjectCache.remove(objectCacheKey);
			}
		}
		else
		{
			if (objectCacheKey == null)
			{
				objectCacheKey = ObjectCache.put(this, timeout);
			}
			else
			{
				ObjectCache.put(objectCacheKey, this, timeout);
			}
		}
	}
	
	/**
	 * Sets a new content with a URL.
	 * The content is stored temporary with the configured timeout.
	 * @param pContent the new content.
	 */
	public void setContent(URL pContent)
	{
		fiContent = null;
		urlContent = pContent;
		
		if (pContent == null)
		{
			if (objectCacheKey != null)
			{
				ObjectCache.remove(objectCacheKey);
			}
		}
		else
		{
			if (objectCacheKey == null)
			{
				objectCacheKey = ObjectCache.put(this, timeout);
			}
			else
			{
				ObjectCache.put(objectCacheKey, this, timeout);
			}
		}
	}
	
	/**
	 * {@inheritDoc}
	 */
	protected void finalize() throws Throwable
	{
		synchronized (this)
		{
			notifyAll();
		}
		try
		{
			fiTempOut.delete();
		}
		catch (Throwable pThrowable)
		{
			// Can do nothing!
		}
		super.finalize();
	}
	
}	// RemoteFileHandle
