/*
 * Copyright 2009 SIB Visions GmbH
 * 
 * Licensed under the Apache License, Version 2.0 (the "License"); you may not
 * use this file except in compliance with the License. You may obtain a copy of
 * the License at
 * 
 * http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 * License for the specific language governing permissions and limitations under
 * the License.
 *
 *
 * History
 *
 * 01.10.2008 - [JR] - creation
 * 01.02.2009 - [JR] - createConnectionProperties: set compression property
 * 04.02.2009 - [JR] - reopen implemented
 *                   - bIgnoreCallError used
 *                   - getProperties implemented
 * 08.04.2009 - [JR] - getProperties: don't mask the password  
 * 23.05.2009 - [JR] - set/getLifeCycleName implemented     
 * 07.07.2009 - [JR] - executeWithSessionContext implemented
 *                   - get(String) implemented    
 * 23.02.2010 - [JR] - #18: setTimeout: property is now numeric and not String
 * 16.09.2011 - [JR] - #23: call/open/close events
 * 07.03.2012 - [JR] - #556
 *                     * set decimal format symbols as connection properties
 *                     * set locale information as connection properties
 *                     * set environment properties as connection properties
 */
package javax.rad.remote;

import java.text.DecimalFormatSymbols;
import java.util.Hashtable;
import java.util.Locale;
import java.util.Map;
import java.util.Map.Entry;

import javax.rad.IPackageSetup;
import javax.rad.remote.event.CallErrorEvent;
import javax.rad.remote.event.CallEvent;
import javax.rad.remote.event.ConnectionEvent;
import javax.rad.remote.event.ICallBackListener;
import javax.rad.remote.event.IConnectionListener;
import javax.rad.remote.event.PropertyEvent;

import com.sibvisions.util.ArrayUtil;
import com.sibvisions.util.ChangedHashtable;

/**
 * The <code>AbstractConnection</code> encapsulates the access to an
 * <code>IConnection</code> implementation.<br>It implements the connection
 * listener handling and an implementation of {@link ConnectionInfo}.
 * 
 * @author Ren Jahn
 * @see IConnection
 */
public abstract class AbstractConnection
{
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Class members
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/** the relevant system properties. */
	private static final String[] USED_SYSPROPS = {"user.name", "os.name", "os.version", "os.arch",
        										   "java.vendor", "java.version", "java.class.version",
        										   "java.vm.name", "file.encoding", "file.separator", 
        										   "path.separator", "line.separator", "user.language",
        										   "user.timezone"};
	
	
	/** the connection implementation. */
	protected IConnection connection;
	
	/** the connection information.*/
	protected ConnectionInfo coninf;
	
	/** all registered connection listeners. */
	private ArrayUtil<IConnectionListener> auConListener = null;
	
	/** internal flag to ignore call errors and do not notify listeners. */
	boolean bIgnoreCallError = false;
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Initialization
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * Creates a new instance of <code>AppliationConnection</code> with an
	 * <code>IConnection</code> implementation.
	 * 
	 * @param pConnection the <code>IConnection</code> implementation
	 */
	protected AbstractConnection(IConnection pConnection)
	{
		this.connection = pConnection;
		
		coninf = new ConnectionInfo(createConnectionProperties()); 
	}
	
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Abstract methods
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * Opens the connection.
	 * 
	 * @throws Throwable if the connection can not be or is already opened
	 */
	public abstract void open() throws Throwable;

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// Overwritten methods
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~

	/**
	 * {@inheritDoc}
	 */
	@Override
	protected void finalize() throws Throwable
	{
		//avoid fireCallError if the connection is not open
		auConListener = null;
		
		close();
		
		super.finalize();
	}

	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	// User-defined methods
	//~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~
	
	/**
	 * Sets the name of the server-side life-cycle object.
	 * 
	 * @param pName the name of the life-cycle object
	 */
	public void setLifeCycleName(String pName)
	{
		coninf.getProperties().put(IConnectionConstants.LIFECYCLENAME, pName);
	}
	
	/**
	 * Gets the name of the server-side life-cycle object.
	 * 
	 * @return the name of the life-cycle object
	 */
	public String getLifeCycleName()
	{
		return (String)coninf.getProperties().get(IConnectionConstants.LIFECYCLENAME);
	}

	/**
	 * Sets the application name which will be used from the remote server
	 * to authenticate the connection, when the connection is not open.
	 * 
	 * @param pApplicationName the application name
	 */
	public void setApplicationName(String pApplicationName)
	{
		if (!isOpen())
		{
			coninf.getProperties().put(IConnectionConstants.APPLICATION, pApplicationName);
		}
	}
	
	/**
	 * Gets the application name which will be used from the remote server
	 * to authenticate the connection.
	 * 
	 * @return the application name
	 */
	public String getApplicationName()
	{
		return (String)coninf.getProperties().get(IConnectionConstants.APPLICATION);
	}
	
	
	/**
	 * Sets the user name which will be used from the remote server
	 * to authenticate the connection, when the connection is not open.
	 * 
	 * @param pUserName the user name
	 */
	public void setUserName(String pUserName)
	{
		coninf.getProperties().put(IConnectionConstants.USERNAME, pUserName);
	}

	/**
	 * Gets the user name which will be used from the remote server
	 * to authenticate the connection.
	 * 
	 * @return the user name
	 */
	public String getUserName()
	{
		return (String)coninf.getProperties().get(IConnectionConstants.USERNAME);
	}
	
	/**
	 * Sets the password which will be used from the remote server
	 * to authenticate the connection, when the connection is not open.
	 * 
	 * @param pPassword the password
	 */
	public void setPassword(String pPassword)
	{
		if (!isOpen())
		{
			coninf.getProperties().put(IConnectionConstants.PASSWORD, pPassword);
		}
	}
	
	/**
	 * Gets the password which will be used from the remote server
	 * to authenticate the connection.
	 * 
	 * @return the password
	 */
	public String getPassword()
	{
		return (String)coninf.getProperties().get(IConnectionConstants.PASSWORD);
	}

	/**
	 * Checks if the connection is open.
	 * 
	 * @return <code>true</code> if the connection is open, otherwise <code>false</code>
	 */
	public boolean isOpen()
	{
		return connection.isOpen(coninf);
	}
	
	/**
	 * Close the connection.
	 * 
	 * @throws Throwable if it is not possible to close the connection
	 */
	public void close() throws Throwable
	{
		//it's sufficient to close the master because all sub connections will be closed
		//automatically on the server
		try
		{
			connection.close(coninf);
			
			fireClose();
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}	

	/**
	 * Reopens the connection.
	 * 
	 * @throws Throwable if an error occurs while opening the connection
	 */
    public void reopen() throws Throwable
    {
    	bIgnoreCallError = true;
    	
    	try
    	{
    		connection.reopen(coninf);
    		
    		fireReOpen();
    	}
    	catch (Throwable th)
    	{
    		fireCallError(th);
    		
    		throw th;
    	}
    	finally
    	{
    		bIgnoreCallError = false;
    	}
    }

	/**
	 * Returns the connection to the server.
	 * 
	 * @return the connection
	 */
	public IConnection getConnection()
	{
		return connection;
	}
	
	/**
	 * Gets the connection identifier for this connection.
	 * 
	 * @return the connection identifier
	 */
	public Object getConnectionId()
	{
		return coninf.getConnectionId();
	}
	
	/**
	 * Calls desired methods of server objects, through the connection.
	 *
	 * @param pCallBack callback listeners for asynchronous or null for synchronous calls 
	 * @param pObjectName list of already mapped server object names/aliases
	 * @param pMethod method names which should be called
	 * @param pParams parameters for the method calls
	 * @return result list from the remote method calls
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object[] call(ICallBackListener[] pCallBack, String[] pObjectName, String[] pMethod, Object[][] pParams) throws Throwable
	{
		try
		{
			Object[] obj = connection.call(coninf, pObjectName, pMethod, pParams, pCallBack);
			
			if (auConListener != null && obj != null)
			{
				for (int i = 0; i < obj.length; i++)
				{
					fireCall(pCallBack != null && pCallBack.length > i ? pCallBack[i] : null,
							 pObjectName != null && pObjectName.length > i ? pObjectName[i] : null,
							 pMethod[i],
							 pParams != null && pParams.length > i ? pParams[i] : null);
				}
			}
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}	
	
	/**
	 * The method will call a remote method from a remote object. 
	 * 
	 * @param pCallBack callback listener for asynchronous or null for synchronous call 
	 * @param pObjectName an already mapped server object name/alias
	 * @param pMethod method name which should be called
	 * @param pParams parameters for the method call
	 * @return result from the remote method call
	 * @throws Throwable communication error, security checks, invalid method, ...
	 */
	public Object call(ICallBackListener pCallBack, String pObjectName, String pMethod, Object... pParams) throws Throwable
	{
		try
		{
			Object obj = connection.call(coninf, new String[] {pObjectName}, new String[] {pMethod}, new Object[][] {pParams}, new ICallBackListener[] {pCallBack})[0];
			
			fireCall(pCallBack, pObjectName, pMethod, pParams);
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * The method will call a remote method from a remote object. 
	 * 
	 * @param pCallBack callback listener for asynchronous or null for synchronous call 
	 * @param pObjectName an already mapped server object name/alias
	 * @param pMethod method name which should be called
	 * @return result from the remote method call
	 * @throws Throwable communication error, security checks, invalid method, ...
	 */
	public Object call(ICallBackListener pCallBack, String pObjectName, String pMethod) throws Throwable
	{
		try
		{
			Object obj = connection.call(coninf, new String[] {pObjectName}, new String[] {pMethod}, null, new ICallBackListener[] {pCallBack})[0];
			
			fireCall(pCallBack, pObjectName, pMethod);

			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}

	/**
	 * Calls desired methods of server objects, through the connection. This call will be
	 * synchronous.
	 *
	 * @param pObjectName list of already mapped server object names/aliases
	 * @param pMethod method names which should be called
	 * @param pParams parameters for the method calls
	 * @return result list from the remote method calls
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object[] call(String[] pObjectName, String[] pMethod, Object[][] pParams) throws Throwable
	{
		try
		{
			Object[] obj = connection.call(coninf, pObjectName, pMethod, pParams, null);
			
			if (auConListener != null && obj != null)
			{
				for (int i = 0; i < obj.length; i++)
				{
					fireCall(null,
							 pObjectName != null && pObjectName.length > i ? pObjectName[i] : null,
							 pMethod[i],
							 pParams != null && pParams.length > i ? pParams[i] : null);
				}
			}
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}	

	/**
	 * Calls desired methods of server objects, through the connection. This call will be
	 * synchronous.
	 *
	 * @param pObjectName list of already mapped server object names/aliases
	 * @param pMethod method names which should be called
	 * @return result list from the remote method calls
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object[] call(String[] pObjectName, String[] pMethod) throws Throwable
	{
		try
		{
			Object[] obj = connection.call(coninf, pObjectName, pMethod, null, null);
			
			if (auConListener != null && obj != null)
			{
				for (int i = 0; i < obj.length; i++)
				{
					fireCall(null,
							 pObjectName != null && pObjectName.length > i ? pObjectName[i] : null,
							 pMethod[i]);
				}
			}
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}	

	/**
	 * The method will call a remote method from a remote object. The call will be 
	 * synchronous. 
	 * 
	 * @param pObjectName an already mapped server object name/alias
	 * @param pMethod method name which should be called
	 * @param pParams parameters for the method call
	 * @return result from the remote method call
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object call(String pObjectName, String pMethod, Object... pParams) throws Throwable
	{
		try
		{
			Object obj = connection.call(coninf, new String[] {pObjectName}, new String[] {pMethod}, new Object[][] {pParams}, null)[0];
			
			fireCall(null, pObjectName, pMethod, pParams);
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}

	/**
	 * The method will call a remote method from a remote object. The call will be 
	 * synchronous.
	 * 
	 * @param pObjectName an already mapped server object name/alias
	 * @param pMethod method name which should be called
	 * @return result from the remote method call
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object call(String pObjectName, String pMethod) throws Throwable
	{
		try
		{
			Object obj = connection.call(coninf, new String[] {pObjectName}, new String[] {pMethod}, null, null)[0];
			
			fireCall(null, pObjectName, pMethod);
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * The method will call a remote action. The call will be 
	 * synchronous.
	 * 
	 * @param pAction action which should be called
	 * @return result from the remote action call
	 * @throws Throwable communication error, security checks, invalid action, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object callAction(String pAction) throws Throwable
	{
		try
		{
			Object obj = connection.call(coninf, null, new String[] {pAction}, null, null)[0];
			
			fireCallAction(null, pAction);
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * The method will call a remote action. The call will be 
	 * synchronous.
	 * 
	 * @param pAction action which should be called
	 * @param pParams parameters for the action call
	 * @return result from the remote action call
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object callAction(String pAction, Object... pParams) throws Throwable
	{
		try
		{
			Object obj = connection.call(coninf, null, new String[] {pAction}, new Object[][] {pParams}, null)[0];
			
			fireCallAction(null, pAction, pParams);
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}

	/**
	 * The method will call one or more remote actions. The call will be synchronous. 
	 * 
	 * @param pAction list of actions which should be called
	 * @param pParams parameters for the action calls
	 * @return result list from the remote action calls
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object[] callAction(String[] pAction, Object[][] pParams) throws Throwable
	{
		try
		{
			Object[] obj = connection.call(coninf, null, pAction, pParams, null);
			
			if (obj != null)
			{
				for (int i = 0; i < obj.length; i++)
				{
					fireCallAction(null, pAction[i], pParams != null && pParams.length > i ? pParams[i] : null);
				}
			}
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}

	/**
	 * The method will call one or more remote actions. The call will be synchronous. 
	 * 
	 * @param pAction list of actions which should be called
	 * @return result list from the remote action calls
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object[] callAction(String[] pAction) throws Throwable
	{
		try
		{
			Object[] obj = connection.call(coninf, null, pAction, null, null);
			
			if (obj != null)
			{
				for (int i = 0; i < obj.length; i++)
				{
					fireCallAction(null, pAction[i]);
				}
			}
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}

	/**
	 * The method will call a remote action.
	 * 
	 * @param pCallBack callback listener for asynchronous or null for synchronous call 
	 * @param pAction action which should be called
	 * @return result from the remote action call
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object callAction(ICallBackListener pCallBack, String pAction) throws Throwable
	{
		try
		{
			Object obj = connection.call(coninf, null, new String[] {pAction}, null, new ICallBackListener[] {pCallBack})[0];
			
			fireCallAction(pCallBack, pAction);
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * The method will call a remote action. 
	 * 
	 * @param pCallBack callback listener for asynchronous or null for synchronous call 
	 * @param pAction action which should be called
	 * @param pParams parameters for the action call
	 * @return result from the remote action call
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object callAction(ICallBackListener pCallBack, String pAction, Object... pParams) throws Throwable
	{
		try
		{
			Object obj = connection.call(coninf, null, new String[] {pAction}, new Object[][] {pParams}, new ICallBackListener[] {pCallBack})[0];
			
			fireCallAction(pCallBack, pAction, pParams);

			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * The method will call one or more remote actions.
	 * 
	 * @param pCallBack callback listeners for asynchronous or null for synchronous calls 
	 * @param pAction list of actions which should be called
	 * @param pParams parameters for the action calls
	 * @return result list from the remote action calls
	 * @throws Throwable communication error, security checks, invalid method, ...
	 * @see IConnection#call(ConnectionInfo, String[], String[], Object[][], ICallBackListener[])
	 */
	public Object[] callAction(ICallBackListener[] pCallBack, String[] pAction, Object[][] pParams) throws Throwable
	{
		try
		{
			Object[] obj = connection.call(coninf, null, pAction, pParams, pCallBack);
			
			if (obj != null)
			{
				for (int i = 0; i < obj.length; i++)
				{
					fireCallAction(pCallBack != null && pCallBack.length > i ? pCallBack[i] : null,
							       pAction[i],
							       pParams != null && pParams.length > i ? pParams[i] : null);
				}
			}
			
			return obj;
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * Sets a connection property. It's not allowed to set/change client properties.
	 * 
	 * @param pName the property name
	 * @param pValue the value of the property or <code>null</code> to delete the property
	 * @throws Throwable communication error, security checks, invalid method, ...
	 */
	public void setProperty(String pName, Object pValue) throws Throwable
	{
		try
		{
			Object oOld = connection.getProperty(coninf, pName);
			
			connection.setProperty(coninf, pName, pValue);
			
			firePropertyChanged(pName, oOld, pValue);
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * Gets a connection property.
	 * 
	 * @param pName the property name
	 * @return the value of the property or null if the property is not available
	 * @throws Throwable communication error, security checks, invalid method, ...
	 */	
	public Object getProperty(String pName) throws Throwable
	{
		try
		{
			return connection.getProperty(coninf, pName);
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * Gets a copy of the current properties.
	 * 
	 * @return the known properties and its values
	 * @throws Throwable communication error, security checks, invalid method, ...
	 */
	public Hashtable<String, Object> getProperties() throws Throwable
	{
		try
		{
			return connection.getProperties(coninf);
		}
		catch (Throwable th)
		{
			fireCallError(th);
			
			throw th;
		}
	}
	
	/**
	 * Sets the timeout of the connection.
	 * 
	 * @param pMinutes the timeout in minutes
	 * @throws Throwable communication error, security checks, invalid method, ...
	 */
	public void setTimeout(int pMinutes) throws Throwable
	{
		setProperty(IConnectionConstants.PREFIX_SERVER + IConnectionConstants.PREFIX_SESSION + "timeout", Integer.valueOf(pMinutes));
	}
	
    /**
     * Gets the properties which will be sent to the server when opening a new connection.
     * 
     * @return the properties as list with key/value pairs
     */
    protected ChangedHashtable<String, Object> createConnectionProperties()
    {
    	ChangedHashtable<String, Object> chtProperties = new ChangedHashtable<String, Object>();
		
		//-----------------------------------------------------------
		// System Properties
		//-----------------------------------------------------------

		String sValue;
		String sPrefix = IConnectionConstants.PREFIX_CLIENT + IConnectionConstants.PREFIX_SYSPROP;
		
		for (int i = 0, anz = AbstractConnection.USED_SYSPROPS.length; i < anz; i++)
		{
			try
			{
				sValue = System.getProperty(AbstractConnection.USED_SYSPROPS[i]);
				
				if (sValue != null)
				{
					chtProperties.put(sPrefix + USED_SYSPROPS[i], sValue);
				}
			}
			catch (Exception e)
			{
				//e.g. SecurityException in an Applet
			}
		}
		
		sPrefix = IConnectionConstants.PREFIX_CLIENT + IConnectionConstants.PREFIX_ENVPROP;
		
		try
		{
			Map<String, String> mapEnv = System.getenv();
			
			String sKey;
			
			for (Entry<String, String> entry : mapEnv.entrySet())
			{
				sKey = entry.getKey();
				sValue = entry.getValue();
			
				if (sKey != null && sValue != null)
				{
					chtProperties.put(sPrefix + entry.getKey(), sValue);
				}
			}
		}
		catch (Exception e)
		{
			//e.g. SecurityException
		}

		DecimalFormatSymbols dfs = new DecimalFormatSymbols();
		
		sPrefix = IConnectionConstants.PREFIX_CLIENT + "decimalFormatSymbols.";
		
		chtProperties.put(sPrefix + "currencySymbol", dfs.getCurrencySymbol());
		chtProperties.put(sPrefix + "currencyCode", dfs.getCurrency().getCurrencyCode());
		chtProperties.put(sPrefix + "internationalCurrencySymbol", dfs.getInternationalCurrencySymbol());
		chtProperties.put(sPrefix + "decimalsSparator", Character.toString(dfs.getDecimalSeparator()));
		chtProperties.put(sPrefix + "digit", Character.toString(dfs.getDigit()));
		chtProperties.put(sPrefix + "groupingSeparator", Character.toString(dfs.getGroupingSeparator()));
		chtProperties.put(sPrefix + "infinity", dfs.getInfinity());
		chtProperties.put(sPrefix + "minusSign", Character.toString(dfs.getMinusSign()));
		chtProperties.put(sPrefix + "monetaryDecimalSeparator", Character.toString(dfs.getMonetaryDecimalSeparator()));
		chtProperties.put(sPrefix + "NaN", dfs.getNaN());
		chtProperties.put(sPrefix + "patternSeparator", Character.toString(dfs.getPatternSeparator()));
		chtProperties.put(sPrefix + "percent", Character.toString(dfs.getPercent()));
		chtProperties.put(sPrefix + "perMill", Character.toString(dfs.getPerMill()));
		chtProperties.put(sPrefix + "zeroDigit", Character.toString(dfs.getZeroDigit()));
		
		Locale loc = Locale.getDefault();
		
		chtProperties.put(IConnectionConstants.PREFIX_CLIENT + "locale.country", loc.getCountry());
		chtProperties.put(IConnectionConstants.PREFIX_CLIENT + "locale.language", loc.getLanguage());
		chtProperties.put(IConnectionConstants.PREFIX_CLIENT + "locale.variant", loc.getVariant());
		
		//-----------------------------------------------------------
		// Client Information
		//-----------------------------------------------------------

		chtProperties.put(IConnectionConstants.PREFIX_CLIENT + "oslib_version", IPackageSetup.JVX_OS_VERSION);
		chtProperties.put(IConnectionConstants.PREFIX_CLIENT + "spec_version", IPackageSetup.SPEC_VERSION);
		
		//-----------------------------------------------------------
		// Communication Properties
		//-----------------------------------------------------------
		
		chtProperties.put(IConnectionConstants.COMPRESSION, "true");
		
		return chtProperties;
    }    
    
    /**
     * Adds an {@link IConnectionListener} to the list of registered listeners.
     * 
     * @param pListener the new connection listener
     */
    public void addConnectionListener(IConnectionListener pListener)
    {
    	if (auConListener == null)
    	{
    		auConListener = new ArrayUtil<IConnectionListener>();
    	}
    	
    	auConListener.add(pListener);
    }
    
    /**
     * Removes an {@link IConnectionListener} from the list of registered listeners.
     * 
     * @param pListener the new connection listener
     */
    public void removeConnectionListener(IConnectionListener pListener)
    {
    	if (auConListener != null)
    	{
    		auConListener.remove(pListener);
    		
    		if (auConListener.size() == 0)
    		{
    			auConListener = null;
    		}
    	}
    }

    /**
     * Gets all registered {@link IConnectionListener}s.
     * 
     * @return an array with {@link IConnectionListener}s or null if no listener is registered
     */
    public IConnectionListener[] getConnectionListener()
    {
    	if (auConListener != null)
    	{
    		IConnectionListener[] listener = new IConnectionListener[auConListener.size()];
    		
    		return auConListener.toArray(listener);
    	}
    	
    	return null;
    }
    
    /**
     * Fires the <code>callError</code> methods to all registered connection listeners.
     * 
     * @param pError the occured error
     */
    protected void fireCallError(Throwable pError)
    {
    	//ignore call errors if necessary
    	if (bIgnoreCallError)
    	{
    		return;
    	}
    	
    	if (auConListener != null)
    	{
    		CallErrorEvent event = new CallErrorEvent(this, pError);
    		
    		for (int i = 0, anz = auConListener.size(); i < anz; i++)
    		{
    			auConListener.get(i).callError(event);
    		}
    	}
    }

    /**
     * Fires the <code>connectionOpened</code> methods to all registered connection listeners.
     */
    protected void fireOpen()
    {    	
    	if (auConListener != null)
    	{
    		ConnectionEvent event = new ConnectionEvent(this);
    		
    		for (int i = 0, anz = auConListener.size(); i < anz; i++)
    		{
    			auConListener.get(i).connectionOpened(event);
    		}    		
    	}
    }
    
    /**
     * Fires the <code>connectionReOpened</code> methods to all registered connection listeners.
     */
    protected void fireReOpen()
    {
    	if (auConListener != null)
    	{
    		ConnectionEvent event = new ConnectionEvent(this);
    		
    		for (int i = 0, anz = auConListener.size(); i < anz; i++)
    		{
    			auConListener.get(i).connectionReOpened(event);
    		}    		
    	}
    }

    /**
     * Fires the <code>connectionClosed</code> methods to all registered connection listeners.
     */
    protected void fireClose()
    {
    	if (auConListener != null)
    	{
    		ConnectionEvent event = new ConnectionEvent(this);
    		
    		for (int i = 0, anz = auConListener.size(); i < anz; i++)
    		{
    			auConListener.get(i).connectionClosed(event);
    		}    		
    	}
    }
    
    /**
     * Fires the <code>objectCalled</code> methods to all registered connection listeners.
     * 
     * @param pCallBackListener the callback listener
     * @param pObjectName the name of the remote object
     * @param pMethodName the name of the method
     * @param pParams the method parameters
     */
    protected void fireCall(ICallBackListener pCallBackListener, String pObjectName, String pMethodName, Object... pParams)
    {
    	if (auConListener != null)
    	{
    		CallEvent event = new CallEvent(pObjectName, pMethodName, pParams, pCallBackListener != null);
    		
    		for (int i = 0, anz = auConListener.size(); i < anz; i++)
    		{
    			auConListener.get(i).objectCalled(event);
    		}    		
    	}
    }

    /**
     * Fires the <code>actionCalled</code> methods to all registered connection listeners.
     * 
     * @param pCallBackListener the callback listener
     * @param pAction the name of the action
     * @param pParams the method parameters
     */
    protected void fireCallAction(ICallBackListener pCallBackListener, String pAction, Object... pParams)
    {
    	if (auConListener != null)
    	{
    		CallEvent event = new CallEvent(null, pAction, pParams, pCallBackListener != null);
    		
    		for (int i = 0, anz = auConListener.size(); i < anz; i++)
    		{
    			auConListener.get(i).actionCalled(event);
    		}    		
    	}
    }
    
    /**
     * Fires the <code>propertyChanged</code> methods to all registered connection listeners.
     * 
     * @param pName the property name
     * @param pOldValue the old value
     * @param pNewValue the new value
     */
    protected void firePropertyChanged(String pName, Object pOldValue, Object pNewValue)
    {
    	if (auConListener != null)
    	{
    		PropertyEvent event = new PropertyEvent(pName, pOldValue, pNewValue);
    		
    		for (int i = 0, anz = auConListener.size(); i < anz; i++)
    		{
    			auConListener.get(i).propertyChanged(event);
    		}    		
    	}
    }
    
}	// AbstractConnection
